#include "ds50daq/DAQ/V1495Driver.hh"

#include "artdaq/DAQdata/Debug.hh"
#include "artdaq/DAQdata/GeneratorMacros.hh"
#include "cetlib/exception.h"
#include "ds50daq/DAQ/Config.hh"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "detail/caen_io.hh"

#include <string>
#include <vector>

using fhicl::ParameterSet;
using ds50::V1495Fragment;

namespace V1495 {
  enum V1495_registers {
    fw_revision = 0x100C,
    port_a_mask = 0x1010,
    port_b_mask = 0x1014,
    run_control = 0x1018,
    window_inibit = 0x101C,
    mf_extention = 0x1020,
    trigger_control = 0x1024,
    run_number = 0x1028,
    test_pulser_dpm = 0x102C,
    status = 0x1030,
    test_pulser_read = 0x1034,
    irq_latch = 0x1080,
    irq_level = 0x8004,
    irq_id = 0x8006,
    reset_module = 0x800A,
    serial_msb = 0x8180,
    serial_lsb = 0x8184
  };
}


ds50::V1495Driver::V1495Driver(ParameterSet const & ps): DS50FragmentGenerator(ps.get<ParameterSet> ("generator_ds50")) {
  if (ps.get<bool> ("reset_vme", false)) {
    init_link (ps);
    caen_io_->vme_sysreset ();
    close_link ();
    usleep (1000000);
  }

  init_link (ps);

  init_metadata (ps);

  init_trigger (ps);

  init_buffers (ps);

  dont_stop_ = ps.get<bool> ("dont_stop", false);
}

ds50::V1495Driver::~V1495Driver() noexcept {
  close_link ();
}

void ds50::V1495Driver::close_link () {
  caen_io_ = 0;
}


void ds50::V1495Driver::init_link (ParameterSet const& ps) {
  caen_io_ = std::unique_ptr<detail::caen_io>(new detail::caen_io (cvA32_U_DATA, cvD32, ps.get<uint32_t> ("vme_base"), ps.get<std::string> ("link_type"), ps.get<int> ("link_number"), ps.get<int> ("link_node", 0)));

  caen_io_->write_register (V1495::reset_module, 1);
  usleep (100000);

  caen_io_->write_register (V1495::irq_level, 1);

  irqwait_ms_ = ps.get<int> ("irqwait_ms", 200);
  irqwait_checks_ = ps.get<int> ("irqwait_checks", 2);
}

void ds50::V1495Driver::init_metadata (ParameterSet const&) {
  metadata_.serial = ((caen_io_->read_register (V1495::serial_msb, cvD16) & 0xff) << 8) | (caen_io_->read_register (V1495::serial_lsb, cvD16) & 0xff);
  metadata_.firmware_version = caen_io_->read_register (V1495::fw_revision);
}

void ds50::V1495Driver::init_trigger (ParameterSet const& ps) {
  bridge_run_enable_ = ps.get<bool> ("bridge_run_enable", false);
  if (bridge_run_enable_) caen_io_->enable_outputs ();

  std::vector<int> masked_channels = ps.get <std::vector<int>> ("masked_channels", std::vector<int>(0));
  uint64_t mask = 0;
  for (int ch: masked_channels) mask |= 1 << ch;
  if (mask) {
    caen_io_->write_register (V1495::port_a_mask, ~uint32_t(mask));
    caen_io_->write_register (V1495::port_b_mask, ~uint32_t(mask >> 32) & 0xFF);
  }

  float acquisition_gate_us = ps.get<float> ("acquisition_gate_us");
  caen_io_->write_register (V1495::window_inibit, int(acquisition_gate_us / 20e-3));
  caen_io_->write_register (V1495::run_control, 2);

  float memory_full_extension_us = ps.get<float> ("memory_full_extension_us", 0);
  if (memory_full_extension_us > 0) {
    caen_io_->set_register_bits (V1495::run_control, 4);
    caen_io_->write_register (V1495::mf_extention, int(memory_full_extension_us / 20e-3));
  }

  caen_io_->write_register (V1495::trigger_control, 2 | (39 << 26) | (5 << 20) | (5 << 16));  // some valid default for the board

  int pulser_frequency = ps.get<int> ("pulser_frequency");
  if (pulser_frequency > 0) {
    caen_io_->set_register_bits (V1495::run_control, (pulser_frequency << 16));
    caen_io_->set_register_bits (V1495::trigger_control, 1 << 1);
  } else caen_io_->clear_register_bits (V1495::trigger_control, 1 << 1);

  int low_threshold = ps.get<int> ("low_threshold") & 0x3f;
  if (low_threshold > 0) {
    int high_threshold = ps.get<int> ("high_threshold", 39) & 0x3f;
    int majority_time_10ns = ps.get<int> ("majority_time_10ns", 5) & 0xf;

    caen_io_->set_register_bits (V1495::trigger_control, 1 | (majority_time_10ns << 16) | (low_threshold << 20) | (high_threshold << 26));
  }

#ifdef WRITE_DPM
  for (uint8_t i = 0; i < 8; i++) {
    uint32_t reg = ((uint16_t(1) << (i + 1)) - 1);
    caen_io_->write_register (V1495::test_pulser_dpm, reg << 16);
    caen_io_->write_register (V1495::test_pulser_dpm, (i * 0x100) << 16 | 1);
  }
  caen_io_->write_register (V1495::test_pulser_dpm, 0);
  caen_io_->set_register_bits (V1495::trigger_control, 1 << 10);
#endif
}

void ds50::V1495Driver::init_buffers (ParameterSet const&) {
}

void ds50::V1495Driver::start_ () {
  if (bridge_run_enable_) caen_io_->set_outputs ();

  caen_io_->write_register (V1495::run_number, run_number ());

  pause_ ();
  caen_io_->set_register_bits (V1495::run_control, 1);
  
  usleep (1000);
  resume_ ();
}

void ds50::V1495Driver::stop_ () {
  if (bridge_run_enable_) caen_io_->clear_outputs ();

  if (!dont_stop_) caen_io_->clear_register_bits (V1495::run_control, 1);
}

void ds50::V1495Driver::pause_ () {
  caen_io_->set_register_bits (V1495::run_control, 1 << 4);
}

void ds50::V1495Driver::resume_ () {
  caen_io_->clear_register_bits (V1495::run_control, 1 << 4);
}

void ds50::V1495Driver::shutdown_ () {
  caen_io_->vme_sysreset ();
}

bool ds50::V1495Driver::getNext__(artdaq::FragmentPtrs& frags) {
  bool will_stop = should_stop ();

  for (int j = 0; j < irqwait_checks_; j++) {
    if (!will_stop && !(caen_io_->read_register (V1495::status) & 0x10) && !caen_io_->wait_irq (1, irqwait_ms_)) continue;
    //if (!caen_io_->wait_irq (1, irqwait_ms_)) continue;

    if (gettimeofday (&metadata_.trigger_time, nullptr) < 0) mf::LogError ("V1495") << "gettimeofday failed";

    std::unique_ptr<artdaq::Fragment> frag(new artdaq::Fragment(std::ceil (V1495Fragment::pod_size () / sizeof (artdaq::RawDataType)), ev_counter (), fragment_id (), Config::V1495_FRAGMENT_TYPE, metadata_));

    int read_size = caen_io_->read_block (0, &*frag->dataBegin (), V1495Fragment::pod_size () / sizeof (uint32_t), false);
    if (read_size <= 0) continue;

    caen_io_->write_register (V1495::irq_latch, 1);


    V1495Fragment const v1495_frag(*frag.get ());
    //std::cout << std::hex << std::showbase << v1495_frag << std::endl;

    int ev_counter_diff = v1495_frag.trigger_counter () - ev_counter ();
    if (ev_counter_diff) {
      mf::LogError ("V1495") << "board " << fragment_id () << " jump " << ev_counter_diff << "(" << v1495_frag.trigger_counter () << " vs " << ev_counter () << ") at event " << ev_counter ();

      for (int i = 0; ev_counter_diff < 100 && i < ev_counter_diff; ++i)
        frags.emplace_back(new artdaq::Fragment(ev_counter () + i, fragment_id (), Config::MISSED_FRAGMENT_TYPE));

      ev_counter_inc (ev_counter_diff);
    }

    ev_counter_inc ();
    frags.emplace_back (std::move (frag));
  }

  if (!frags.size () && will_stop) {
    mf::LogDebug("V1495") << "status is " << std::showbase << std::hex << caen_io_->read_register (V1495::status);
    return false;
  }
  return true;
}

DEFINE_ARTDAQ_GENERATOR(ds50::V1495Driver)
